/* PHAVerLite: PHAVer + PPLite.
   Copyright (C) 2019-2020 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PHAVerLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "extended_pplite.hh"

// The abstract class for polyhedra

struct abs_poly {
  abs_poly() = default;
  virtual ~abs_poly() {}

  abs_poly(const abs_poly&) = delete;
  abs_poly(abs_poly&&) = delete;
  abs_poly& operator=(const abs_poly&) = delete;
  abs_poly& operator=(abs_poly&&) = delete;

  virtual abs_poly* clone() const = 0;

  virtual void minimize() const = 0;

  virtual bool is_empty() const = 0;
  virtual bool is_universe() const = 0;
  virtual bool is_bounded() const = 0;
  virtual bool is_topologically_closed() const = 0;
  virtual bool boxed_contains(const abs_poly& y) const = 0;
  virtual bool constrains(Var v) const = 0;
  virtual bool contains(const abs_poly& y) const = 0;
  virtual bool equals(const abs_poly& y) const = 0;
  virtual bool is_disjoint_from(const abs_poly& y) const = 0;

  virtual dim_type space_dim() const = 0;
  virtual size_t affine_dim() const = 0;
  virtual dim_type num_min_cons() const = 0;
  virtual Cons copy_cons() const = 0;
  virtual Gens copy_gens() const = 0;
  virtual Poly_Con_Rel relation_with(const Con& c) const = 0;
  virtual Poly_Gen_Rel relation_with(const Gen& g) const = 0;
  virtual BBox get_bounding_box() const = 0;
  virtual bool min(const Linear_Expr& expr, const Integer& inhomo,
                   Integer& inf_n, Integer& inf_d,
                   bool& minimum, Gen& g) const = 0;
  virtual bool max(const Linear_Expr& expr, const Integer& inhomo,
                   Integer& inf_n, Integer& inf_d,
                   bool& maximum, Gen& g) const = 0;
  virtual void print(std::ostream& os) const = 0;
  virtual size_t get_memory_in_bytes() const = 0;

  // Modifiers
  virtual void m_swap(abs_poly& y) noexcept = 0;
  virtual void set_empty() = 0;
  virtual void add_con(const Con& c) = 0;
  virtual void add_cons(const Cons& cs) = 0;
  virtual void add_gen(const Gen& g) = 0;
  virtual void add_gens(const Gens& gs) = 0;

  virtual void con_hull_assign(const abs_poly& y) = 0;
  virtual void concatenate_assign(const abs_poly& y) = 0;
  virtual void intersection_assign(const abs_poly& y) = 0;
  virtual void poly_hull_assign(const abs_poly& y) = 0;
  virtual void time_elapse_assign(const abs_poly& y) = 0;
  virtual void topological_closure_assign() = 0;
  virtual void widening_assign(const abs_poly& y,
                               widen_type widen_impl) = 0;
  virtual void widening_assign(const abs_poly& y, const Cons& cs,
                               widen_type widen_impl) = 0;
  virtual abs_poly* split(const Con& con, Topol topol) = 0;

  virtual void
  affine_image(Var var, const Linear_Expr& expr, const Integer& inhomo,
               const Integer& den = 1) = 0;
  virtual void
  affine_preimage(Var var, const Linear_Expr& expr, const Integer& inhomo,
                  const Integer& den = 1) = 0;
  virtual void
  parallel_affine_image(const Vars& vars,
                        const Linear_Exprs& exprs,
                        const Integers& inhomos,
                        const Integers& dens) = 0;

  virtual void unconstrain(const var_ref_set& vars) = 0;
  virtual void add_space_dims(dim_type d, bool project = false) = 0;
  virtual void map_space_dims(const Dims& pfunc) = 0;
  virtual void remove_space_dims(const var_ref_set& vars) = 0;
  virtual void remove_higher_space_dims(dim_type new_dim) = 0;

}; // abs_poly

inline void
swap(abs_poly& x, abs_poly& y) noexcept {
  x.m_swap(y);
}

inline std::ostream&
operator<<(std::ostream& os, const abs_poly& x) {
  x.print(os);
  return os;
}

//////////////////////////////////////////////////////////////////////////

// The adapter class template generating concrete classes.

template <typename PH>
struct poly_adapter : public abs_poly {
  using poly_type = PH;
  poly_type ph;

  // Helper
  static const poly_type& get_poly(const abs_poly& y) {
    assert(dynamic_cast<const poly_adapter*>(&y));
    return static_cast<const poly_adapter&>(y).ph;
  }

  explicit poly_adapter(const poly_type& y_ph) : ph(y_ph) {}
  explicit poly_adapter(poly_type&& y_ph) : ph(std::move(y_ph)) {}

  ~poly_adapter() = default;
  poly_adapter(const poly_adapter&) = delete;
  poly_adapter(poly_adapter&&) = delete;
  poly_adapter& operator=(const poly_adapter&) = delete;
  poly_adapter& operator=(poly_adapter&&) = delete;

  // Public interface
  explicit poly_adapter(dim_type d, Spec_Elem spec, Topol t)
    : ph(d, spec, t) {}

  poly_adapter* clone() const override {
    return new poly_adapter(ph);
  }

  void minimize() const override { ph.minimize(); }
  bool is_empty() const override { return ph.is_empty(); }
  bool is_universe() const override { return ph.is_universe(); }
  bool is_bounded() const override { return ph.is_bounded(); }
  bool is_topologically_closed() const override {
    return ph.is_topologically_closed();
  }
  bool boxed_contains(const abs_poly& y) const override {
    return ph.boxed_contains(get_poly(y));
  }
  bool constrains(Var v) const override { return ph.constrains(v); }
  bool contains(const abs_poly& y) const override {
    return ph.contains(get_poly(y));
  }
  bool equals(const abs_poly& y) const override {
    return ph.equals(get_poly(y));
  }
  bool is_disjoint_from(const abs_poly& y) const override {
    return ph.is_disjoint_from(get_poly(y));
  }

  dim_type space_dim() const override { return ph.space_dim(); }
  size_t affine_dim() const override { return ph.affine_dim(); }
  dim_type num_min_cons() const override { return ph.num_min_cons(); }
  Cons copy_cons() const override { return ph.copy_cons(); }
  Gens copy_gens() const override { return ph.copy_gens(); }
  Poly_Con_Rel relation_with(const Con& c) const override {
    return ph.relation_with(c);
  }
  Poly_Gen_Rel relation_with(const Gen& g) const override {
    return ph.relation_with(g);
  }
  BBox get_bounding_box() const override { return ph.get_bounding_box(); }
  bool min(const Linear_Expr& expr, const Integer& inhomo,
           Integer& inf_n, Integer& inf_d,
           bool& minimum, Gen& g) const override {
    return ph.min(expr, inhomo, inf_n, inf_d, minimum, g);
  }
  bool max(const Linear_Expr& expr, const Integer& inhomo,
           Integer& inf_n, Integer& inf_d,
           bool& maximum, Gen& g) const override {
    return ph.max(expr, inhomo, inf_n, inf_d, maximum, g);
  }
  void print(std::ostream& os) const override { ph.print(os); }
  size_t get_memory_in_bytes() const override {
    return pplite::total_memory_in_bytes(ph);
  }

  // Modifiers
  void m_swap(abs_poly& y) noexcept override {
    assert(dynamic_cast<poly_adapter*>(&y));
    ph.m_swap(static_cast<poly_adapter&>(y).ph);
  }
  void set_empty() override { ph.set_empty(); }
  void add_con(const Con& c) override { ph.add_con(c); }
  void add_cons(const Cons& cs) override { ph.add_cons(cs); }
  void add_gen(const Gen& g) override { ph.add_gen(g); }
  void add_gens(const Gens& gs) override { ph.add_gens(gs); }

  void con_hull_assign(const abs_poly& y) override {
    ph.con_hull_assign(get_poly(y));
  }
  void concatenate_assign(const abs_poly& y) override {
    ph.concatenate_assign(get_poly(y));
  }
  void intersection_assign(const abs_poly& y) override {
    ph.intersection_assign(get_poly(y));
  }
  void poly_hull_assign(const abs_poly& y) override {
    ph.poly_hull_assign(get_poly(y));
  }
  void time_elapse_assign(const abs_poly& y) override {
    ph.time_elapse_assign(get_poly(y));
  }
  void topological_closure_assign() override {
    ph.topological_closure_assign();
  }
  void widening_assign(const abs_poly& y, widen_type widen_impl) override {
    ph.widening_assign(get_poly(y), widen_impl);
  }
  void widening_assign(const abs_poly& y, const Cons& cs,
                       widen_type widen_impl) override {
    ph.widening_assign(get_poly(y), cs, widen_impl);
  }
  poly_adapter* split(const Con& con, Topol topol) override {
    return new poly_adapter(ph.split(con, topol));
  }

  void
  affine_image(Var var, const Linear_Expr& expr, const Integer& inhomo,
               const Integer& den = 1) override {
    ph.affine_image(var, expr, inhomo, den);
  }
  void
  affine_preimage(Var var, const Linear_Expr& expr, const Integer& inhomo,
                  const Integer& den = 1) override {
    ph.affine_image(var, expr, inhomo, den);
  }
  void
  parallel_affine_image(const Vars& vars,
                        const Linear_Exprs& exprs,
                        const Integers& inhomos,
                        const Integers& dens) override {
    ph.parallel_affine_image(vars, exprs, inhomos, dens);
  }

  void unconstrain(const var_ref_set& vars) override {
    ph.unconstrain(vars.begin(), vars.end());
  }
  void add_space_dims(dim_type d, bool project) override {
    ph.add_space_dims(d, project);
  }
  void map_space_dims(const Dims& pfunc) override {
    ph.map_space_dims(pfunc);
  }
  void remove_space_dims(const var_ref_set& vars) override {
    ph.remove_space_dims(vars.begin(), vars.end());
  }
  void remove_higher_space_dims(dim_type new_dim) override {
    ph.remove_higher_space_dims(new_dim);
  }

}; // poly_adapter

// Declarations of explicit instantiations of the available concrete classes.
extern template class poly_adapter<pplite::Poly>;
extern template class poly_adapter<pplite::Stats<pplite::Poly>>;
extern template class poly_adapter<pplite::F_Poly>;
extern template class poly_adapter<pplite::Stats<pplite::F_Poly>>;

// A poor man's object factory.
// Note: Stats versions require ad-hoc configuration anyway.
enum class Poly_Kind {
  POLY, F_POLY,
  POLY_STATS, F_POLY_STATS,
  SIZE
};

constexpr auto poly_kind_size = static_cast<int>(Poly_Kind::SIZE);

constexpr const char* poly_kind_names[poly_kind_size] = {
  "Poly", "F_Poly",
  "Poly_Stats", "F_Poly_Stats"
};

extern Poly_Kind poly_kind;

// Returns true (and sets global variable accordingly)
// if "name" is the name of a (supported) concrete class.
bool set_poly_kind(const std::string& name, bool noisy = false);

// Use this to actually build concrete objects
inline abs_poly*
make_poly(dim_type d = 0,
          Spec_Elem spec = Spec_Elem::UNIVERSE,
          Topol t = Topol::NNC) {
  using namespace pplite;
  switch (poly_kind) {
  case Poly_Kind::POLY:
    return new poly_adapter<Poly>(d, spec, t);
  case Poly_Kind::F_POLY:
    return new poly_adapter<F_Poly>(d, spec, t);
  case Poly_Kind::POLY_STATS:
    return new poly_adapter<Stats<Poly>>(d, spec, t);
  case Poly_Kind::F_POLY_STATS:
    return new poly_adapter<Stats<F_Poly>>(d, spec, t);
  default:
    assert(false);
    return nullptr;
  }
}
